@@ -24,6 +24,7 @@
             eval-sizeof-type
             eval-alignof-type
             eval-sizeof-expr
+            eval-typeof-expr
             eval-offsetof
             find-offsets
             find-sizes
@@ -361,43 +362,71 @@
          (loop specl declrs (cdr fields)))
         ((comp-decl ,specl ,declr)
          (loop specl (list declr) (cdr fields)))
-        (,_ (sferr "missed in cxeval,find-fields") #f)))
-     (else (sferr "find-field: missed:\n") (pperr (car fields)) #f))))
+        (,_ (sferr "missed in cxeval(find-field)") #f)))
+     (else (sferr "cxeval(find-field): missed:\n") (pperr (car fields)) #f))))
 
 (define (lookup-type-field udecl name)
   (sx-match udecl
-    ((type-name (decl-spec-list (type-spec (struct-def ,field-list))) ,declr)
+    ((type-name (decl-spec-list (type-spec (struct-def ,field-list))) . ,_)
      (let* ((field-list (clean-field-list field-list)))
        (find-field (sx-tail field-list) name)))
-    ((type-name (decl-spec-list (type-spec (union-def ,field-list))) ,declr)
+    ((type-name (decl-spec-list (type-spec (union-def ,field-list))) . ,_)
      (let* ((field-list (clean-field-list field-list)))
        (find-field (sx-tail field-list) name)))
     (,_ #f)))
 
-(define* (eval-typeof-expr expr #:optional (udict '())) ;; => `(type-name ...
+;; is aggr type decl (not couting typedef
+(define (aggr-decl? adecl)
+  (let* ((specl (sx-ref adecl 1))
+         (tspec (sx-find 'type-spec specl))
+         (tag (sx-ref* tspec 1 1)))
+    (and (memq tag '(struct-def union-def)))))
 
+;; if struct or union ref, expand to struct or union def
+(define (expose-aggr adecl udict)
+  (define (key tag) (case tag ((struct-ref) 'struct) ((union-ref) 'union)))
+  (let* ((specl (sx-ref adecl 1))
+         (declr (sx-ref adecl 2))
+         (tspec (sx-find 'type-spec specl))
+         (tag (sx-tag (sx-ref* tspec 1))))
+    (case tag
+      ((typename)
+       (call-with-values
+           (lambda () (splice-typename specl declr (sx-ref* tspec 1 1) udict))
+         (lambda (specl declr) (sx-list 'type-name #f specl declr))))
+      ((struct-ref union-ref)
+       (let* ((name (sx-ref* tspec 1 1 1))
+              (form (cons (key tag) name))
+              (specl (replace-aggr-ref specl name form udict)))
+         (sx-list 'type-name #f specl declr)))
+      (else adecl))))
+
+;; @deffn {Scheme} eval-typeof-expr expr udict => `(type-name specl declr)
+;; @var{expr} should be @code{(typeof-expr expr)} but is not.
+;; @end deffn
+(define (eval-typeof-expr expr udict)
   ;; a->b.c => a .b ->c ; a in udict, b in a, c in b
   ;; what about vectors?  DO LATER
   ;; a->b[1].c => a .b ->c ; a in udict, b in a, c in b
-  (define* (typeof-next next #:optional type-name)
-    ;; => (type-name specl declr)
-    ;; where type is an aggregate type (I think)
+  (define (typeof-next next)
     (sx-match next
       ((ident ,name)
        (let* ((udecl (assoc-ref udict name))
               (specl (sx-ref udecl 1))
-              (declr (sx-ref udecl 2))
-              (exp `(type-name ,specl ,declr)))
-         exp))
+              (declr (sx-ref udecl 2)))
+         `(type-name ,specl ,declr)))
+      ((cast ,tname ,expr)
+       tname)
       ((p-expr ,expr)
        (typeof-next expr))
       ((i-sel ,ident ,expr)
        (and expr (typeof-next `(d-sel ,ident (de-ref ,expr)))))
       ((d-sel (ident ,name) ,expr)
        (let* ((tname (typeof-next expr))
-              (xdecl (and tname (expand-typerefs tname udict))))
-         ;; TODO: check xdecl's declr
-         (and xdecl (lookup-type-field xdecl name))))
+              (xdecl (and tname (expose-aggr tname udict))))
+         (and xdecl
+              (not (pointer-declr? (sx-ref xdecl 2)))
+              (lookup-type-field xdecl name))))
       ((de-ref ,expr)
        (let ((tname (typeof-next expr)))
          (and tname
@@ -406,6 +435,8 @@
                  `(type-name ,specl (,declr ,expr)))
                 ((,decl ,specl (,declr (ptr-declr (pointer ,rest) ,expr)))
                  `(type-name ,specl (,declr (ptr-declr ,rest) ,expr)))
+                ((,decl ,specl (abs-ptr-declr (pointer)))
+                 `(type-name ,specl)) ; maybe FIXME
                 (,_ #f)))))
        ((array-ref ,indx ,expr)
         (let ((tname (typeof-next expr)))
@@ -416,7 +447,6 @@
                  (,_  #f)))))))
 
   (typeof-next expr))
-(export eval-typeof-expr)
 
 ;; =============================================================================
 
